from argparse import ArgumentParser
from argparse import Namespace

from .io.io import read_from_files_as_dict
from .ondict import ONDict
from .typing import Dict
from .typing import Set


class CLArgs:
    def add_arguments_to_argparse(
        self, parser: ArgumentParser, prefix: str = "", ignore: Set[str] = None
    ):
        """Add config arguments to ArgumentParser.

        The method add to the parser arguments that exist in the default config supplied
        on the instantiation of :class:`~resconfig.ResConfig` object. The long option
        will be “-”-delimited for nested keys, e.g., ``config["foo.bar"]`` will be
        mapped to ``--foo-bar``. If ``prefix`` is supplied, the prefix is prepended to
        the long option, e.g., ``--prefix-foo-bar``.

        To prevent the method from adding arguments to the parser for some keys, supply
        ``ignore`` with a set of keys to ignore, e.g., ``{"foo.bar"}``, so that
        ``--foo-bar`` will not be added.

        Args:
            parser: Parser object to which config arguments are added.
            prefix: Argument prefix.
            ignore: Config keys to ignore.
        """
        ignore = ignore or set()
        for key in self._default.allkeys():
            if ".".join(key) not in ignore:
                default = self._default[key]
                key = (prefix,) + key if prefix else key
                longopt = "--" + "-".join(k.replace(r"\.", ".") for k in key)
                parser.add_argument(longopt, default=default)

    def prepare_from_argparse(
        self,
        args: Namespace,
        config_file_arg: str = None,
        prefix: str = "",
        keymap: Dict[str, str] = None,
    ):
        """Prepare config from :class:`~argparse.ArgumentParser` result.

        See the docstring of :meth:`~resconfig.ResConfig.add_arguments_to_argparse` for
        the rule on how parser long options map to config keys. If long options do not
        directly map to config keys by that rule, you can supply ``prefix`` or
        ``keymap`` to define your own mapping. For example, if you want to map
        ``--long-opt`` (which would be parsed as the ``long_opt`` attribute) to the
        ``foo.bar`` config key, use ``{"long_opt": "foo.bar"}`` for ``keymap``.

        If you want to allow config file(s) to be specified through
        :class:`~argparse.ArgumentParser`, specify the attribute name for the argument
        to ``config_file_arg``.

        Args:
            args: Parse result returned from
                :meth:`~argparse.ArgumentParser.parse_args`.
            config_file_arg: The attribute name for config files.
            prefix: Argument prefix.
            keymap: Key mapping from the parsed argument name to the config key.
        """
        args = vars(args)
        keymap = keymap or {}
        conf = ONDict()
        if config_file_arg and config_file_arg in args:
            paths = args[config_file_arg]
            if paths:
                paths = paths if isinstance(paths, (list, tuple)) else [paths]
                conf.merge(read_from_files_as_dict(paths, self._merge_config_files))

        # TODO: Prune items not in default config above?

        for k, v in (item for item in args.items() if item[0] != config_file_arg):
            if k in keymap:
                k = keymap[k]
            else:
                k = ".".join(i.replace(".", r"\.") for i in k.split("_"))
                if prefix and k.startswith(prefix):
                    k = k[len(prefix) + 1 :]
            if k in self._default:
                conf.merge(ONDict({k: v}))

        self._clargs = conf



from functools import wraps

from .actions import Action
from .ondict import ONDict
from .ondict import isdict
from .typing import Any
from .typing import Key
from .typing import List
from .typing import Optional
from .typing import WatchFunction


class Watchers(ONDict):
    _create = True
    __watcher_key = "__watchers__"

    def deregister(self, key: Key, func: Optional[WatchFunction] = None):
        """Deregister the watch function for the key."""
        ref = self.get(key)
        if not ref:
            raise KeyError(f"Watch function not registered for {key}")
        if func is None:
            if self.__watcher_key in ref:
                del ref[self.__watcher_key]
        else:
            if self.__watcher_key not in ref:
                raise KeyError(f"Watch functions not registered for {key}")
            try:
                ref[self.__watcher_key].remove(func)
            except ValueError:
                raise ValueError(f"{func!r} not registered for {key}")
            # If this was the last watch function, remove the node.
            if not ref[self.__watcher_key]:
                del ref[self.__watcher_key]

    def register(self, key: Key, func: WatchFunction):
        """Register the watch function for the key."""
        self.setdefault(key, self.__class__()).setdefault(
            self.__watcher_key, []
        ).append(func)

    def exists(self, key: Key) -> bool:
        """Test if any watch function exists for the key."""
        return key in self and self.__watcher_key in self[key]

    def funcs(self, key: Key) -> List[WatchFunction]:
        """Get the list of watch functions for the key."""
        if key in self:
            return self[key].get(self.__watcher_key, [])
        return []

    def trigger(self, key: Key, action: Action, oldval: Any, newval: Any):
        """Trigger the watch functions for the key."""
        for func in self.funcs(key):
            func(action, oldval, newval)


class Watchable:
    """Mix-in for adding the watch functionality."""

    def deregister(self, key: Key, func: Optional[WatchFunction] = None):
        """Deregister the watch function for the key."""
        self._watchers.deregister(key, func)

    def register(self, key: Key, func: WatchFunction):
        """Register the watch function for the key."""
        self._watchers.register(key, func)

    def __reload(self, key, d):
        _key = key[-1]
        if not isdict(d[_key]):
            self._watchers.trigger(key[1:], Action.RELOADED, d[_key], d[_key])
            return
        for subkey, v in d[_key].items():
            self.__reload(key + (subkey,), d[_key])
        self._watchers.trigger(key[1:], Action.RELOADED, d[_key], d[_key])

    def reload(self):
        """Trigger all watch functions using the current configuration.

        Note that the watch functions for the keys that do not exist in the current
        configuration will not be triggered.
        """
        # The reason why the recursive visit is on the conf, not the watchers is that we
        # want to trigger functions in order of configuration.
        self.__reload(("__ROOT__",), {"__ROOT__": self._conf})

    def watch(self, key: Key) -> WatchFunction:
        """Decorate a function to make it a watch function for the key."""

        def deco(f):
            @wraps(f)
            def _deco(*args, **kwargs):
                return f(*args, **kwargs)

            self.register(key, _deco)
            return _deco

        return deco



from copy import deepcopy
from enum import Enum
from pathlib import Path

from ..ondict import ONDict
from ..typing import FilePath
from ..typing import List
from ..typing import NewType
from ..typing import Optional
from ..utils import experimental
from . import ini
from . import json
from . import toml
from . import yaml
from .utils import ensure_path


class FileType(Enum):
    ini = 1
    json = 2
    toml = 3
    yaml = 4


FileTypes = NewType("FileTypes", FileType)


def _suffix_to_filetype(filename: FilePath) -> FileType:
    filetype = {
        ".ini": FileType.ini,
        ".json": FileType.json,
        ".toml": FileType.toml,
        ".yaml": FileType.yaml,
        ".yml": FileType.yaml,
    }.get(Path(filename).suffix, FileType.ini)
    return filetype


def _read_as_dict(filename: FilePath, filetype: Optional[FileTypes] = None) -> ONDict:
    if filetype is None:
        filetype = _suffix_to_filetype(filename)
    load = {
        FileType.ini: ini.load,
        FileType.json: json.load,
        FileType.toml: toml.load,
        FileType.yaml: yaml.load,
    }.get(filetype, ini.load)
    with open(filename) as f:
        loaded = load(f)
    return ONDict(loaded)


def read_from_files_as_dict(paths: List[FilePath], merge: bool = False) -> ONDict:
    """Read the config from a file(s) as an :class:`ONDict`.

    How the config is constructed depends on the ``merge`` flag. If :obj:`True`, the
    files are searched in the reverse order, and the configs are read from all existing
    files and merged in that order. If :obj:`False`, then the files are searched for in
    the order given in ``paths``, and the first existing file provides the config to be
    read (and the rest are ignored).

    Args:
        paths: A list of config file paths.
        merge: The flag for the merge mode; see the function description.

    Returns:
        An :class:`~resconfig.ondict.ONDict` object.
    """
    d = ONDict()
    paths = reversed(paths) if merge else paths
    for path in paths:
        path = ensure_path(path)
        if path.is_file():
            content = _read_as_dict(path)
            d.merge(content)
            if not merge:
                break
    return d


class IO:
    """The mix-in to add file IO functionality."""

    def update_from_files(self, paths: List[FilePath], merge=False):
        """Update config from files.

        How the config is constructed depends on the ``merge`` flag. If :obj:`True`, the
        files are searched in the reverse order, and the configs are read from all
        existing files and merged in that order. If :obj:`False`, then the files are
        searched for in the order given in ``paths``, and the first existing file
        provides the config to be read (and the rest are ignored).

        Args:
            paths: A list of config file paths.
            merge: The flag for the merge mode; see the function description.
        """
        self.update(read_from_files_as_dict(paths, merge))

    def __update_from_file(self, filename: FilePath, filetype: FileType):
        self.update(_read_as_dict(filename, filetype))

    def update_from_file(self, filename: FilePath):
        """Update config from the file.

        The file type is inferred from the filename extension.
        """
        self.__update_from_file(filename, _suffix_to_filetype(filename))

    def update_from_ini(self, filename: FilePath):
        """Update config from the INI file."""
        self.__update_from_file(filename, FileType.ini)

    def update_from_json(self, filename: FilePath):
        """Update config from the JSON file."""
        self.__update_from_file(filename, FileType.json)

    def update_from_toml(self, filename: FilePath):
        """Update config from the TOML file."""
        self.__update_from_file(filename, FileType.toml)

    def update_from_yaml(self, filename: FilePath):
        """Update config from the YAML file."""
        self.__update_from_file(filename, FileType.yaml)

    def __save(self, filename: FilePath, filetype: Optional[FileType] = None):
        dump = {
            FileType.ini: ini.dump,
            FileType.json: json.dump,
            FileType.toml: toml.dump,
            FileType.yaml: yaml.dump,
        }.get(filetype, ini.dump)

        d = deepcopy(self._conf)
        self._schema.unapply_all(d)

        with open(filename, "w") as f:
            dump(d, f)

    @experimental
    def save_to_file(self, filename: FilePath):
        """Save config to the file.

        The file type is inferred from the filename extension.
        """
        self.__save(filename, _suffix_to_filetype(filename))

    @experimental
    def save_to_ini(self, filename: FilePath):
        """Save config to the INI file."""
        self.__save(filename, FileType.ini)

    @experimental
    def save_to_json(self, filename: FilePath):
        """Save config to the JSON file."""
        self.__save(filename, FileType.json)

    @experimental
    def save_to_toml(self, filename: FilePath):
        """Save config to the TOML file."""
        self.__save(filename, FileType.toml)

    @experimental
    def save_to_yaml(self, filename: FilePath):
        """Save config to the YAML file."""
        self.__save(filename, FileType.yaml)



from functools import wraps


def experimental(f):
    """Note the decorated function or method is experimental."""

    @wraps(f)
    def func(*args, **kwargs):
        return f(*args, **kwargs)

    func.__doc__ = "*Experimental:* " + func.__doc__
    return func



from pathlib import Path
from typing import IO  # noqa
from typing import Callable  # noqa
from typing import Dict  # noqa
from typing import Generator  # noqa
from typing import Iterable  # noqa
from typing import List  # noqa
from typing import NewType  # noqa
from typing import Optional  # noqa
from typing import Set  # noqa
from typing import Text  # noqa
from typing import Tuple  # noqa
from typing import TypeVar  # noqa
from typing import Union

from .actions import Action

from typing import Any  # noqa; noqa


Key = Union[str, Tuple[str]]

WatchFunction = Callable[[Action, Any, Any], None]
"""Callback function, i.e., watcher, that triggers on an event happening at the key. It
takes in Action value, old, and new values.
"""

FilePath = Union[str, Path]

RT = TypeVar("RT")



from .ondict import ONDict


class Schema(ONDict):
    def _cast(self, cls, value):
        if callable(cls):
            try:
                return cls(value)
            except Exception:
                raise TypeError(f"{value!r} cannot be converted to {cls}")
        return value

    def apply(self, key, value):
        item = self.get(key, None)
        cls = item[0] if isinstance(item, tuple) else item
        return self._cast(cls, value)

    def unapply(self, key, value):
        item = self.get(key, None)
        cls = item[1] if isinstance(item, tuple) else item
        return self._cast(cls, value)

    def unapply_all(self, d):
        for key in self.allkeys():
            print(key, "in", d)
            if key in d:
                print("applying")
                d[key] = self.unapply(key, d[key])



from collections.abc import MutableMapping
from configparser import ConfigParser

from ..typing import IO


def load(f: IO) -> dict:
    parser = ConfigParser()
    parser.read_file(f)
    dic = {}
    for section in parser.sections():
        ref = dic.setdefault(_escape_dot(section), {})
        for option in parser[section]:
            ref[_escape_dot(option)] = parser[section][option]
    return dic


def dump(content: dict, f: IO):
    if _depth(content) > 2:
        raise ValueError("INI config does not allow nested options")

    _unescape_all_keys(content)

    parser = ConfigParser()
    parser.read_dict(content)
    parser.write(f)


def _escape_dot(key: str) -> str:
    return key.replace(".", r"\.")


def _unescape_dot(key: str) -> str:
    return key.replace(r"\.", ".")


def _unescape_all_keys(d: dict):
    if not isinstance(d, MutableMapping):
        return
    for k in list(d.keys()):
        _unescape_all_keys(d[k])
        uk = _unescape_dot(k)
        if uk != k:
            d[uk] = d[k]
            del d[k]


def _depth(d: dict, depth: int = 0) -> int:
    if not isinstance(d, MutableMapping):
        return depth
    max_depth = depth
    for key in d:
        max_depth = max(max_depth, _depth(d[key], depth + 1))
    return max_depth



try:
    import yaml
except ImportError:
    yaml = None

from ..ondict import ONDict


def dump(data, stream, **kwargs):
    class _Dumper(yaml.Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items()
        )

    _Dumper.add_representer(ONDict, _dict_representer)

    yaml.dump(data, stream, _Dumper, **kwargs)


def load(stream, **kwargs):
    return yaml.load(stream, yaml.FullLoader, **kwargs)



"""resconfig Configuration Library
==================================

:copyright: (c) 2020 by Taro Sato.
:license: Apache 2.0, see LICENSE.txt for more details.
"""

from .actions import Action  # noqa
from .resconfig import ResConfig  # noqa

__author__ = "Taro Sato"
__author_email__ = "okomestudio@gmail.com"
__version__ = "20.4.2a"
__license__ = "Apache-2.0"



from enum import Enum


class Action(Enum):
    """Action performed by the update method.

    The watch function receives this value as its first argument to change its behavior
    based on the action.
    """

    ADDED = 1
    """The item at the key has been added."""

    MODIFIED = 2
    """The item at the key has been modified."""

    REMOVED = 3
    """The item at the key has been removed."""

    RELOADED = 4
    """The item at the key has been reloaded."""



#!/usr/bin/env python
import codecs
import os
import re

from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def fread(filename):
    with codecs.open(os.path.join(here, filename), "r", encoding="utf-8") as f:
        return f.read()


def meta(category, fpath="src/resconfig/__init__.py"):
    package_root_file = fread(fpath)
    matched = re.search(
        r"^__{}__\s+=\s+['\"]([^'\"]*)['\"]".format(category), package_root_file, re.M
    )
    if matched:
        return matched.group(1)
    raise Exception("Meta info string for {} undefined".format(category))


version = meta("version")
author = meta("author")
author_email = meta("author_email")
license = meta("license")
readme = fread("README.rst")


requires = []
toml_requires = ["toml>=0.10.0"]
yaml_requires = ["PyYAML>=5.3.1"]

setup_requires = ["pytest-runner>=5.2"]

dev_requires = (
    [
        "black>=19.10b0",
        "flake8>=3.7.9",
        "isort[pyproject]>=4.3.21",
        "pre-commit>=2.2.0",
        "seed-isort-config>=2.1.1",
    ]
    + toml_requires
    + yaml_requires
)

doc_requires = ["sphinx>=3.0.1", "sphinx_autodoc_typehints>=1.10.3"]

tests_require = (
    ["coverage[toml]>=5.0.4", "pytest>=5.4.1", "pytest-cov>=2.8.1"]
    + toml_requires
    + yaml_requires
)


setup(
    name="resconfig",
    version=version,
    description="A minimalistic application configuration library for Python",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author=author,
    author_email=author_email,
    url="https://github.com/okomestudio/resconfig",
    platforms=["Linux"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Installation/Setup",
        "Topic :: Utilities",
    ],
    packages=find_packages("src"),
    package_data={"": ["LICENSE"]},
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    license=license,
    scripts=[],
    install_requires=requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    extras_require={
        "dev": dev_requires + tests_require,
        "doc": doc_requires,
        "tests": tests_require,
        "toml": toml_requires,
        "yaml": yaml_requires,
    },
    entry_points={"console_scripts": []},
)



try:
    from toml import dumps as _dump
    from toml import loads as _load
except ImportError:
    _dump = _load = None


def load(f):
    return _load(f.read())


def dump(content, f):
    f.write(_dump(dict(content)))



import os
from collections.abc import Iterable
from copy import deepcopy
from enum import Enum
from logging import getLogger

from .actions import Action
from .clargs import CLArgs
from .io import IO
from .io.io import read_from_files_as_dict
from .io.utils import ensure_path
from .ondict import ONDict
from .ondict import flexdictargs
from .ondict import isdict
from .ondict import merge
from .schema import Schema
from .typing import Any
from .typing import Dict
from .typing import FilePath
from .typing import Key
from .typing import List
from .typing import Optional
from .typing import Tuple
from .typing import WatchFunction
from .watchers import Watchable
from .watchers import Watchers

log = getLogger(__name__)


class Flag(Enum):
    MISSING = 1
    """Flag missing value."""

    REMOVE = 2
    """Flag config item to be removed."""


class ResConfig(Watchable, IO, CLArgs):
    """An application resource configuration.

    This object holds the state and the internal data structure for the configuration.

    Args:
        default: Default config.
        config_files: List of config filename paths.
        envvar_prefix: Prefix used for environment variables used as configuration.
        load_on_init: :obj:`True` to load config on instantiation, :obj:`False` to skip.
        merge_config_files: :obj:`True` to merge all configs from existing files,
            :obj:`False` to read only the config from the first existing file.
        schema: *Experimental:* Config schema.
        watchers: Config watchers.
    """

    def __init__(
        self,
        default: Optional[dict] = None,
        config_files: Optional[List[FilePath]] = None,
        envvar_prefix: str = "",
        load_on_init: bool = True,
        merge_config_files: bool = True,
        schema: Optional[dict] = None,
        watchers: Optional[Dict[Key, List[WatchFunction]]] = None,
    ):
        self._default = ONDict(default or {})
        self._config_files = (
            [ensure_path(p) for p in config_files] if config_files else []
        )
        self._envvar_prefix = envvar_prefix
        self._clargs = ONDict()
        self._merge_config_files = merge_config_files
        self._schema = Schema(schema or {})
        self._watchers = Watchers()
        for k, v in (watchers or {}).items():
            for func in v if isinstance(v, Iterable) else [v]:
                self.register(k, v)

        # This is where the active config is stored.
        self._conf = ONDict()

        if load_on_init:
            self.load()

    def __contains__(self, key):
        return key in self._conf

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(key)
        return self.get(key)

    def _asdict(self) -> dict:
        """Return the config as a dict object."""
        return dict(deepcopy(self._conf))

    def _prepare_config(self) -> ONDict:
        """Prepare a new :class:`ONDict` object with the current object state.

        Returns:
             An :class:`~resconfig.ondict.ONDict` object.
        """
        new = deepcopy(self._default)

        if self._config_files:
            new.merge(
                read_from_files_as_dict(self._config_files, self._merge_config_files)
            )

        env = os.environ
        clargs = self._clargs
        for key in self._default.allkeys():
            envkey = self._envvar_prefix + "_".join(
                k.replace(r"\.", "_").upper() for k in key
            )
            if envkey in env:
                new[key] = env[envkey]

            if key in clargs:
                new[key] = clargs[key]

        return new

    def get(self, key: Key, default: Optional[Any] = None) -> Any:
        """Return the config value for key if it exists, else default.

        Args:
            key: Config key.
            default: Default value if key is not in config.

        Returns:
            The value found for the key.
        """
        try:
            value = self._conf[key]
        except Exception:
            return default
        return deepcopy(value)

    def load(self):
        """Load the prepared config."""
        self.replace(self._prepare_config())

    def __update(
        self, key: Tuple[str], conf: dict, newconf: dict, replace: bool = False
    ) -> Tuple[Action, Any, Any]:
        """Perform config update recursively.

        Tuple of keys to the current node ``k`` holds the full path from the root, i.e.,
        ``("root", "k1", "k2", ..., "k")``, where ``conf["k"]`` and ``newconf["k"]`` is
        the node to be inspected at the current call stack. The full key path needs to
        be retained this way to notify watch functions.

        Args:
            key: Tuple of keys to the current node.
            conf: Old config to be updated.
            newconf: New config to update with.
            replace: :obj:`True` to perform replacement instead of merge.

        Returns:
            A tuple of action, old value, and new value for the current key.
        """
        _key = key[-1]

        if not isdict(newconf[_key]):
            newval = self._schema.apply(key[1:], newconf[_key])
            action = None
            if not isdict(conf):
                oldval = Flag.MISSING
                if newval is not Flag.REMOVE:
                    action = Action.ADDED
            elif _key not in conf or conf[_key] is Flag.MISSING or not conf[_key]:
                oldval = Flag.MISSING
                if newval is not Flag.REMOVE:
                    action = Action.ADDED
            else:
                oldval = deepcopy(conf[_key])
                if newval is Flag.REMOVE:
                    action = Action.REMOVED
                elif oldval != newval:
                    action = Action.MODIFIED
            return action, oldval, newval

        oldval_at_dict_node = deepcopy(conf[_key]) if _key in conf else Flag.MISSING
        newval_at_dict_node = ONDict()

        conf.setdefault(_key, ONDict())

        for subkey in newconf[_key].keys():
            if not isdict(conf[_key]):
                conf[_key] = ONDict()

            action, oldval, newval = self.__update(
                key + (subkey,), conf[_key], newconf[_key], replace=replace
            )

            # Actually update the config storage
            if action in (Action.MODIFIED, Action.ADDED):
                if isdict(newval):
                    newval = merge(conf[_key][subkey], newval)
                    newval_at_dict_node[subkey] = merge(
                        newval_at_dict_node.setdefault(subkey, ONDict()), newval
                    )
                else:
                    newval_at_dict_node[subkey] = newval

                conf[_key][subkey] = newval

            elif action in (Action.REMOVED,):
                del conf[_key][subkey]
                if subkey in newval_at_dict_node:
                    del newval_at_dict_node[subkey]

            # If an action occurs, trigger its watch functions
            if action is not None and self._watchers.exists(key[1:] + (subkey,)):
                self._watchers.trigger(key[1:] + (subkey,), action, oldval, newval)

        if replace:
            seen = set(newconf[_key].keys())
            for subkey in (k for k in list(conf[_key].keys()) if k not in seen):
                if self._watchers.exists(key[1:] + (subkey,)):
                    self._watchers.trigger(
                        key[1:] + (subkey,),
                        Action.REMOVED,
                        conf[_key][subkey],
                        Flag.REMOVE,
                    )
                del conf[_key][subkey]

        # Define the action performed on this dict node.
        action = None
        if newval_at_dict_node:
            if not oldval_at_dict_node or oldval_at_dict_node is Flag.MISSING:
                action = Action.ADDED
                oldval_at_dict_node = Flag.MISSING
            else:
                if newval_at_dict_node != oldval_at_dict_node:
                    action = Action.MODIFIED
        elif not oldval_at_dict_node:
            action = Action.REMOVED

        return action, oldval_at_dict_node, newval_at_dict_node

    @flexdictargs
    def update(self, conf: dict):
        """Perform update of config.

        Args:
            conf: Config to update with.
        """
        k = "__ROOT__"  # Insert a layer for the first iteration
        self.__update((k,), {k: self._conf}, {k: conf})

    @flexdictargs
    def replace(self, conf: dict):
        """Perform replacement of config.

        Args:
            conf: Config for replacement.
        """
        k = "__ROOT__"  # Insert a layer for the first iteration
        self.__update((k,), {k: self._conf}, {k: conf}, replace=True)

    def reset(self):
        """Reset config to default."""
        self.replace(deepcopy(self._default))



import re
from collections import OrderedDict
from collections.abc import MutableMapping
from functools import wraps

from .typing import RT
from .typing import Any
from .typing import Callable
from .typing import Generator
from .typing import Iterable
from .typing import Key
from .typing import Tuple

_default = object()


class ONDict(OrderedDict):
    _create = None

    def __init__(self, *args, **kwargs):
        args, kwargs = _expand_args(args, kwargs)
        super().__init__(*args, **kwargs)

    def __repr__(self):
        items = []
        for k, v in self.items():
            items.append(f"{k!r}: {v!r}")
        return "{{" + ", ".join(items) + "}}"

    def __contains__(self, key):
        try:
            ref, lastkey = _get(self, key, self._create)
        except Exception:
            return False
        return (super() if ref is self else ref).__contains__(lastkey)

    def __delitem__(self, key):
        ref, lastkey = _get(self, key, self._create)
        if lastkey not in ref:
            raise _key_error(ref, key)
        return (super() if ref is self else ref).__delitem__(lastkey)

    def __getitem__(self, key):
        ref, lastkey = _get(self, key, self._create)
        if ref is self:
            return super().__getitem__(lastkey)
        else:
            if lastkey not in ref:
                raise _key_error(ref, key)
            return ref.__getitem__(lastkey)

    def __setitem__(self, key, value):
        ref, lastkey = _get(self, key, self._create)
        return (super() if ref is self else ref).__setitem__(lastkey, value)

    def get(self, key, default=None):
        try:
            ref, lastkey = _get(self, key, self._create)
        except Exception:
            return default
        return (super() if ref is self else ref).get(lastkey, default)

    def pop(self, key, d=_default):
        if d is _default:
            ref, lastkey = _get(self, key, self._create)
            if lastkey not in ref:
                raise _key_error(ref, key)
            return (super() if ref is self else ref).pop(lastkey)
        else:
            try:
                ref, lastkey = _get(self, key, self._create)
            except Exception:
                return d
            return (super() if ref is self else ref).pop(lastkey, d)

    def setdefault(self, key, default=None):
        ref, lastkey = _get(self, key, self._create)
        return (super() if ref is self else ref).setdefault(lastkey, default)

    def update(self, *args, **kwargs):
        args, kwargs = _expand_args(args, kwargs)
        if args:
            if len(args) != 1:
                raise TypeError(f"update expected at most 1 argument, got {len(args)}")
            for k, v in args[0].items():
                self[k] = v
        for k in kwargs:
            self[k] = kwargs[k]

    @classmethod
    def fromkeys(cls, iterable, value=None):
        dic = cls()
        for key in iterable:
            dic = merge(dic, normalize({key: value}))
        return dic

    # custom utility methods

    def __allkeys(self, key, d):
        _key = key[-1]
        if not isdict(d[_key]):
            yield key[1:]
            return
        for k in d[_key]:
            for i in self.__allkeys(key + (k,), d[_key]):
                yield i

    def allkeys(self, as_str=False):
        """Generate all keys to the leaves"""
        for key in self.__allkeys(("__ROOT__",), {"__ROOT__": self}):
            yield ".".join(key) if as_str else key

    def merge(self, d):
        merge(self, d)


def _expand_args(args, kwargs):
    if args:
        arg = args[0]
        if hasattr(arg, "keys"):
            new = normalize(arg)
        else:
            new = ONDict()
            for key, val in arg:
                new = merge(new, normalize({key: val}))
        args = [new] + list(args[1:])
    return args, normalize(kwargs)


def _key_error(obj, key):
    return KeyError(f"'{key}'")


def _type_error(obj, key):
    return TypeError(f"'{type(obj)}' object at '{key}' is not subscriptable")


def _get(dic: dict, key: Key, create: bool = False) -> Tuple[dict, Key]:
    keys = list(normkey(key))
    ref = dic
    for idx, k in enumerate(keys[:-1]):
        try:
            ref = ref[k]
        except KeyError:
            if not create:
                raise _key_error(ref, ".".join(keys[: idx + 1]))
            ref[k] = dic.__class__()
            ref = ref[k]
        except TypeError:
            raise _type_error(ref, ".".join(keys[: idx + 1]))
    if not isdict(ref):
        raise _type_error(ref, key)
    return ref, keys[-1]


def get(dic: dict, key: Key) -> Any:
    ref, key = _get(dic, key)
    return ref[key]


def normkey(key: Key) -> Generator[str, None, None]:
    """Normalize key.

    If the input key is “.”-delimit nested keys, it is split on the delimiter. If the
    input key is a tuple of keys, they are simply yielded.

    In the string form, escaped “.” characters are not interpreted as delimiters.

    Args:
        key: Input key.

    Returns:
        Generator of str keys.

    Raises:
        TypeError: When input key is neither :class:`str` or :class:`tuple`.
    """
    if isinstance(key, tuple):
        for k in key:
            yield k
    elif isinstance(key, str):
        if "." in key:
            for k in re.split(r"(?<!\\)\.", key):
                yield k
        else:
            yield key
    else:
        raise TypeError("key must be str or tuple")


def __merge(a, b):
    for k in b:
        if k in a:
            if isdict(a[k]):
                if isdict(b[k]):
                    __merge(a[k], b[k])
                else:
                    if a[k] != b[k]:
                        a[k] = b[k]
            else:
                if a[k] != b[k]:
                    a[k] = b[k]
        else:
            a[k] = b[k]
    return a


def merge(a: dict, b: dict) -> dict:
    """Merge dict b into a recursively.

    If either dict has leaves that are non-dicts, the leaf in dict b overwrites that in
    dict a.
    """
    return __merge(a, b)


def normalize(d: dict) -> dict:
    """Normalize the :class:`dict` by expanding nested keys.

    This function returns a new :class:`dict` object by expanding the nested keys and
    their values into nested :class:`dict` objects.

    Raises:
        TypeError: When an attempt is made to convert a non-:class:`dict` node to a
            :class:`dict` node.
    """
    if not isinstance(d, MutableMapping):
        return d
    new = d.__class__()
    for key, value in d.items():
        keys = list(normkey(key))
        if len(keys) == 1:
            k = keys[0]
            expanded = normalize(value)
            new[k] = (
                merge(new[k], expanded)
                if k in new and isinstance(new[k], MutableMapping)
                else expanded
            )
        else:
            ref = new
            for idx, k in enumerate(keys[:-1]):
                if k not in ref:
                    ref[k] = d.__class__()
                ref = ref[k]
                if not isinstance(ref, MutableMapping):
                    k = ".".join(keys[: idx + 1])
                    raise TypeError(
                        f"cannot convert a node from non-dict to dict at '{k}'"
                    )
            k = keys[-1]
            expanded = normalize(value)
            ref[k] = (
                merge(ref[k], expanded)
                if k in ref and isinstance(ref[k], MutableMapping)
                else expanded
            )
    return new


def flexdictargs(func: Callable[[dict], RT]) -> Callable[[Iterable, Any], RT]:
    """Decorate a method to add dict.update()-like interface.

    The decorated method should take one :class:`dict` object as an argument. Then the
    decorator makes it possible to accept an iterable or :class:`dict` as keyword
    arguments. The output :class:`dict`
    """

    @wraps(func)
    def f(self, *args, **kwargs):
        if args and isinstance(args[0], MutableMapping):
            d = args[0]
        elif kwargs:
            d = kwargs
        else:
            raise TypeError("invalid input arguments")
        return func(self, normalize(d))

    return f


def isdict(val: Any) -> bool:
    """Test if the value is dict."""
    return isinstance(val, MutableMapping)
